"""Data preprocessing utilities."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

if TYPE_CHECKING:
    from xarray import Dataset


def aggregate_to_daily(ds: Dataset) -> Dataset:
    """Aggregate data to daily values.

    Parameters
    ----------
    ds : Dataset
        Data to aggregate.

    Returns
    -------
    Dataset
        Daily aggregated data.
    """
    daily = ds.groupby(ds.time.dt.date).mean()
    # the groupy changes the coordinate name and type,
    # so we convert it back
    daily["date"] = pd.to_datetime(daily["date"])
    daily = daily.rename({"date": "time"})

    return daily


def time_substitution(ds: Dataset, interval: slice) -> Dataset:
    """Create a counterfactual by substituting data from an interval.

    Takes the data interval and repreats it to fill the whole dataset.
    For the dataset [1, 2, 3, 4, 5], `substitution(ds, slice(0, 0))`
    returns [1, 1, 1, 1, 1].

    Parameters
    ----------
    ds : Dataset
        Data to substitute.
    interval : slice
        Interval to substitute.

    Returns
    -------
    Dataset
        Data with substituted data.
    """
    out = ds.copy()
    ds_slice = ds.sel(time=interval)

    n_d = ds["time"].shape[0]
    n_s = ds_slice["time"].shape[0]

    rep = np.ceil(n_d / n_s)

    # for variable in dataset tile the slice and keep the first n_d values
    for var in ds.data_vars:
        out[var].values = np.tile(ds_slice[var].values, int(rep))[:n_d]

    return out



import gpytorch
import numpy as np
import torch
from discontinuum.engines.gpytorch import MarginalGPyTorch, NoOpMean

from gpytorch.kernels import (
    MaternKernel,
    RBFKernel,
    RQKernel,
    ScaleKernel,
)
from gpytorch.priors import (
    GammaPrior,
    HalfNormalPrior,
    NormalPrior,
)

from linear_operator.operators import MatmulLinearOperator
from rating_gp.models.base import RatingDataMixin, ModelConfig
from rating_gp.plot import RatingPlotMixin
from rating_gp.models.kernels import StageTimeKernel, SigmoidKernel


class PowerLawTransform(torch.nn.Module):
    """
    """
    def __init__(self):
        super(PowerLawTransform, self).__init__()
        self.a = torch.nn.Parameter(torch.rand(1))
        self.b = torch.nn.Parameter(torch.rand(1))
        self.c = torch.nn.Parameter(torch.zeros(1))

    def forward(self, x):
        self.c.data = torch.clamp(self.c.data, max=x.min()-1e-6)
        return self.a + (self.b * torch.log(x - self.c))


class RatingGPMarginalGPyTorch(
    RatingDataMixin,
    RatingPlotMixin,
    # comes last b/c it conflics with Mixin
    MarginalGPyTorch,
):
    """
    Gaussian Process implementation of the LOAD ESTimation (LOADEST) model

    This model currrently uses the marginal likelihood implementation, which is
    fast but does not account for censored data. Censored data require a slower
    latent variable implementation.
    """
    def __init__(
            self,
            model_config: ModelConfig = ModelConfig(),
    ):
        """ """
        super(MarginalGPyTorch, self).__init__(model_config=model_config)
        self.build_datamanager(model_config)
        

    def build_model(self, X, y, y_unc=None) -> gpytorch.models.ExactGP:
        """Build marginal likelihood version of RatingGP
        """
        # assume a constant measurement error for testing
        if y_unc is not None:
            noise = y_unc
        else:
            noise = 0.1**2 * torch.ones(y.shape[0]).reshape(1, -1)
        # TODO: Fix "GPInputWarning: You have passed data through a 
        # FixedNoiseGaussianLikelihood that did not match the size of the fixed
        # noise, *and* you did not specify noise. This is treated as a no-op."
        self.likelihood = gpytorch.likelihoods.FixedNoiseGaussianLikelihood(
            noise=noise,
            learn_additional_noise=False,
        )

        model = ExactGPModel(X, y, self.likelihood)

        return model


class ExactGPModel(gpytorch.models.ExactGP):
    def __init__(self, train_x, train_y, likelihood):
        super(ExactGPModel, self).__init__(train_x, train_y, likelihood)

        n_d = train_x.shape[1]  # number of dimensions
        assert n_d == 2, "Only two dimensions supported"

        self.dims = np.arange(n_d)
        self.time_dim = [self.dims[0]]
        self.stage_dim = [self.dims[1]]

        self.powerlaw = PowerLawTransform()

        # self.mean_module = gpytorch.means.ConstantMean()
        # self.mean_module = gpytorch.means.LinearMean(input_size=1)
        self.mean_module = NoOpMean()

        # self.covar_module = (
        #     (self.cov_stage() * self.cov_stagetime())
        #     + self.cov_residual()
        # )
      
        # Stage * time kernel with large time length
        # + stage * time kernel only at low stage with smaller time length
        self.covar_module = (
            (self.cov_stage()
             * self.cov_time(ls_prior=GammaPrior(concentration=10, rate=5)))
            + (self.cov_stage()
               * self.cov_time(ls_prior=GammaPrior(concentration=2, rate=5))
               * SigmoidKernel(
                   active_dims=self.stage_dim,
                   # a_prior=NormalPrior(loc=20, scale=1),
                   b_constraint=gpytorch.constraints.Interval(
                       train_x[:, self.stage_dim].min(),
                       train_x[:, self.stage_dim].max()
                   ),
               )
              )
        )

    def forward(self, x):
        self.powerlaw.b.data.clamp_(1.5, 2.5)
        x_t = x.clone()
        x_t[:, self.stage_dim] = self.powerlaw(x_t[:, self.stage_dim])

        mean_x = self.mean_module(x_t[:, self.stage_dim])
        covar_x = self.covar_module(x_t)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)

    def cov_stage(self, ls_prior=None):
        eta = HalfNormalPrior(scale=1)

        return ScaleKernel(
            MaternKernel(
                active_dims=self.stage_dim,
                lengthscale_prior=ls_prior,
            ),
            outputscale_prior=eta,
        )

    def cov_time(self, ls_prior=None):
        eta = HalfNormalPrior(scale=1)

        return ScaleKernel(
            MaternKernel(
                active_dims=self.time_dim,
                lengthscale_prior=ls_prior,
            ),
            outputscale_prior=eta,
        )

    def cov_stagetime(self):
        eta = HalfNormalPrior(scale=1)
        ls = GammaPrior(concentration=2, rate=1)

        return ScaleKernel(
            StageTimeKernel(
                active_dims=self.dims,
                # lengthscale_prior=ls,
            ),
            # outputscale_prior=eta,
        )

    def cov_residual(self):
        eta = HalfNormalPrior(scale=0.2)
        ls = GammaPrior(concentration=2, rate=10)

        return ScaleKernel(
            MaternKernel(
                ard_num_dims=2,
                nu=1.5,
                active_dims=self.dims,
                lengthscale_prior=ls,
            ),
            outputscale_prior=eta,
        )



import gpytorch
import numpy as np
import torch
from linear_operator.operators import MatmulLinearOperator, to_dense


class StageTimeKernel(gpytorch.kernels.Kernel):
    """A time RBF kernel with a stage-variable length scale.

    A scalar length scale is multiplied by the stage, which is taken to a
    variable power, to impose a stage variablity.
    """
    has_lengthscale = True

    def __init__(self,
                 a_prior=None,
                 a_constraint=gpytorch.constraints.Positive(),
                 **kwargs):
        """Initialize the kernel

        Parameters
        ----------
        a_prior : Prior
            The prior to impose on the power variable.
        a_constraint : Constraint
            The constraint to impose on the power variable
        """
        super().__init__(**kwargs)

        # register the raw parameters
        self.register_parameter(
            name='raw_a',
            parameter=torch.nn.Parameter(torch.ones(*self.batch_shape, 1, 1))
        )

        # register the constraints
        # if a_constraint is not None:
        self.register_constraint("raw_a", a_constraint)

        # set the parameter prior, see
        # https://docs.gpytorch.ai/en/latest/module.html#gpytorch.Module.register_prior
        if a_prior is not None:
            self.register_prior(
                "a_prior",
                a_prior,
                lambda m: m.a,
                lambda m, v : m._set_a(v),
            )

    # now set up the 'actual' paramter
    @property
    def a(self):
        # when accessing the parameter, apply the constraint transform
        return self.raw_a_constraint.transform(self.raw_a)

    @a.setter
    def a(self, value):
        return self._set_a(value)

    def _set_a(self, value):
        if not torch.is_tensor(value):
            value = torch.as_tensor(value).to(self.raw_a)
        # when setting the paramater, transform the actual value to a raw one by applying the inverse transform
        self.initialize(raw_a=self.raw_a_constraint.inverse_transform(value))

    # create a RBF Kernel that has a stage dependent scale length
    def forward(self, x1, x2, **params):
        dims = np.arange(x1.shape[1])
        time_dim = [dims[0]]
        stage_dim = [dims[1]]

        # make the scale length for the time dimension dependent on stage
        # shift stage to have a minimum value of 1
        x1_stage = x1[:, stage_dim] - x1[:, stage_dim].min() + 1
        x2_stage = x2[:, stage_dim] - x2[:, stage_dim].min() + 1
        
        x1_stage_dep_time_lengthscale = self.lengthscale * x1_stage ** self.a
        x2_stage_dep_time_lengthscale = self.lengthscale * x2_stage ** self.a

        # apply lengthscale
        x1_ = x1[:, time_dim].div(x1_stage_dep_time_lengthscale)
        x2_ = x2[:, time_dim].div(x2_stage_dep_time_lengthscale)

        # calculate the distance between inputs
        diff = self.covar_dist(x1_, x2_, square_dist=True, **params)
        return diff.div_(-2).exp_()


class PowerLawKernel(gpytorch.kernels.Kernel):
    """Power Law Kernel

    This kernel is the rating curve power law equivalent to the linear kernel.
    The power law equation is given by:

        f(x) = a + (b * ln(x - c))
    """
    def __init__(self,
                 a_prior=None,
                 a_constraint=None,
                 b_prior=None,
                 b_constraint=gpytorch.constraints.Positive(),
                 c_prior=None,
                 c_constraint=gpytorch.constraints.Positive(),
                 **kwargs):
        """Initialize the kernel

        Parameters
        ----------
        a_prior : Prior
            The prior to impose on `a`.
        a_constraint : Constraint
            The constraint to impose on `a`
        b_prior : Prior
            The prior to impose on `b`.
        b_constraint : Constraint
            The constraint to impose on `b`
        c_prior : Prior
            The prior to impose on `c`.
        c_constraint : Constraint
            The constraint to impose on `c`
        """
        super().__init__(**kwargs)

        # register the raw parameters
        self.register_parameter(
            name='raw_a',
            parameter=torch.nn.Parameter(torch.rand(*self.batch_shape, 1, 1))
        )
        b_start = torch.rand(*self.batch_shape, 1, 1)
        self.register_parameter(
            name='raw_b',
            parameter=torch.nn.Parameter(b_start)
        )
        self.register_parameter(
            name='raw_c',
            parameter=torch.nn.Parameter(torch.zeros(*self.batch_shape, 1, 1))
        )

        # register the constraints
        if a_constraint is None:
            a_constraint = gpytorch.constraints.Interval(-10, 10)
        self.register_constraint("raw_a", a_constraint)
        self.register_constraint("raw_b", b_constraint)
        self.register_constraint("raw_c", c_constraint)

        # set the parameter prior
        if a_prior is not None:
            self.register_prior(
                "a_prior",
                a_prior,
                lambda m: m.a,
                lambda m, v : m._set_a(v),
            )
        if b_prior is not None:
            self.register_prior(
                "b_prior",
                b_prior,
                lambda m: m.b,
                lambda m, v : m._set_b(v),
            )
        if c_prior is not None:
            self.register_prior(
                "c_prior",
                c_prior,
                lambda m: m.c,
                lambda m, v : m._set_c(v),
            )

    # set the actual parameters
    @property
    def a(self):
        return self.raw_a_constraint.transform(self.raw_a)

    @a.setter
    def a(self, value):
        return self._set_a(value)

    def _set_a(self, value):
        if not torch.is_tensor(value):
            value = torch.as_tensor(value).to(self.raw_a)
        self.initialize(raw_a=self.raw_a_constraint.inverse_transform(value))

    @property
    def b(self):
        return self.raw_b_constraint.transform(self.raw_b)

    @b.setter
    def b(self, value):
        return self._set_b(value)

    def _set_b(self, value):
        if not torch.is_tensor(value):
            value = torch.as_tensor(value).to(self.raw_b)
        self.initialize(raw_b=self.raw_b_constraint.inverse_transform(value))

    @property
    def c(self):
        return self.raw_c_constraint.transform(self.raw_c)

    @c.setter
    def c(self, value):
        return self._set_c(value)

    def _set_c(self, value):
        if not torch.is_tensor(value):
            value = torch.as_tensor(value).to(self.raw_c)
        self.initialize(raw_c=self.raw_c_constraint.inverse_transform(value))

    # the kernel function
    def forward(self, x1, x2, diag=False, **params):
        x1_ = self.a + (self.b * torch.log(x1 - self.c))
        x2_ = self.a + (self.b * torch.log(x2 - self.c))

        # Calculate cross product
        prod = MatmulLinearOperator(x1_, x2_.transpose(-2, -1))
        if diag:
            return prod.diagonal(dim1=-1, dim2=-2)
        else:
            return prod


class SigmoidKernel(gpytorch.kernels.Kernel):
    """Sigmoid Kernel

    This kernel can be multiplied by another kernel to give a breakpoint in the
    data using a sigmoid

    Note: The strength of the slope of the sigmoid is currently fixed to a steep
    slope. This can be turned back into a parameter, but it cause numerical
    instabilities during fitting.
    """
    def __init__(
        self,
        # a_prior=None,
        # a_constraint=None,
        b_prior=None,
        b_constraint=gpytorch.constraints.Positive(),
        **kwargs,
        ):
        """Initialize the kernel

        Parameters
        ----------
        b_prior : Prior
            The prior to impose on `b`.
        b_constraint : Constraint
            The constraint to impose on `b`
        """
        super().__init__(**kwargs)

        self.a = 20
        # self.register_parameter(
        #     name='raw_a',
        #     parameter=torch.nn.Parameter(torch.zeros(*self.batch_shape, 1, 1))
        # )
        self.register_parameter(
            name='raw_b',
            parameter=torch.nn.Parameter(torch.zeros(*self.batch_shape, 1, 1))
        )

        # register the constraints
        # if a_constraint is None:
        #     a_constraint = gpytorch.constraints.Positive()
        # self.register_constraint("raw_a", a_constraint)
        self.register_constraint("raw_b", b_constraint)

        # set the parameter prior
        # if a_prior is not None:
        #     self.register_prior(
        #         "a_prior",
        #         a_prior,
        #         lambda m: m.a,
        #         lambda m, v : m._set_a(v),
        #     )
        if b_prior is not None:
            self.register_prior(
                "b_prior",
                b_prior,
                lambda m: m.b,
                lambda m, v : m._set_b(v),
            )

    # set the 'actual' paramters
    # @property
    # def a(self):
    #     return self.raw_a_constraint.transform(self.raw_a)

    # @a.setter
    # def a(self, value):
    #     return self._set_a(value)

    # def _set_a(self, value):
    #     if not torch.is_tensor(value):
    #         value = torch.as_tensor(value).to(self.raw_a)
    #     self.initialize(raw_a=self.raw_a_constraint.inverse_transform(value))

    @property
    def b(self):
        return self.raw_b_constraint.transform(self.raw_b)

    @b.setter
    def b(self, value):
        return self._set_b(value)

    def _set_b(self, value):
        if not torch.is_tensor(value):
            value = torch.as_tensor(value).to(self.raw_b)
        self.initialize(raw_b=self.raw_b_constraint.inverse_transform(value))

    def forward(self, x1, x2, last_dim_is_batch=False, diag=False, **params):
        # sigmoid = 1/(1+exp(-a(x-b)))
        # `a` is the sharpness of the slope, larger absolute values = sharper slope
        # the sign of `a` determines which side of the curve is 0 and the other is 1
        # `b` is the offset of of the curve at a sigmoid value of 0.5
        x1_ = 1/(1 + torch.exp(self.a * (x1 - self.b)))
        x2_ = 1/(1 + torch.exp(self.a * (x2 - self.b)))
        
        prod = MatmulLinearOperator(x1_, x2_.transpose(-2, -1))
        if diag:
            return prod.diagonal(dim1=-1, dim2=-2)
        else:
            return prod



import sys
import warnings

__all_gpytorch__ = [
    # Sub-packages in models.gpytorch
    "LoadestGPMarginalGPyTorch",
]

__all_pymc__ = [
    # PyMC sub-packages in models.pymc
    "LoadestGPMarginalPyMC",
]

for name in __all_gpytorch__:
    exec(f"from .models.gpytorch import {name}")

__all__ = __all_gpytorch__.copy()


# check if pymc is installed, then import the sub-packages
try:
    import pymc
    __all__.extend(__all_pymc__)

    for name in __all_pymc__:
        exec(f"from .models.pymc import {name}")

except ModuleNotFoundError:
    # if pymc is not installed, raise a warning
    warnings.warn(
        f"`pymc` is not installed: {', '.join(__all_pymc__)}"
        " will not be available.",
        UserWarning,
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import functools
from abc import ABC, abstractmethod

from discontinuum.data_manager import DataManager
from discontinuum.pipeline import StandardErrorPipeline, StandardPipeline

if TYPE_CHECKING:
    from typing import Dict, Optional
    from xarray import DataArray, Dataset


class BaseModel(ABC):
    def __init__(self, model_config: Optional[Dict] = None):
        """Base class for all models.

        Parameters
        ----------
        model_config : dict, optional
            Configuration for the model. The default is None.
        """
        if model_config is None:
            model_config = {}

        self.model_config = model_config
        self.dm = None
        self.is_fitted = False

    @abstractmethod
    def fit(self,
            covariates: Dataset,
            target: Dataset,
            **kwargs,
            ):
        """Fit model to data.

        Parameters
        ----------
        covariates : Dataset
            Covariates for training.
        target : Dataset
            Target data for training.
        kwargs : dict
            Additional keyword arguments.
        """
        self.is_fitted = True
        return self

    @abstractmethod
    def predict(self, covariates: Dataset) -> DataArray:
        """Use a fitted model to make predictions on new data.

        Parameters
        ----------
        covariates : Dataset
            Covariates for prediction.
        """
        pass

    @abstractmethod
    def build_model(self, X, y):
        """
        TODO Sometimes this sets self.model and sometimes this returns model.
        Not sure that we can standardize the behavior for different engines.
        """
        pass

    @abstractmethod
    def build_datamanager(self):
        """Build DataManager for the model."""
        self.dm = DataManager(
            target_pipeline=StandardPipeline,
            error_pipeline=StandardErrorPipeline,
            covariate_pipelines=StandardPipeline,
        )


def is_fitted(func):
    """Decorator checks whether model has been fit."""

    @functools.wraps(func)
    def inner(self, *args, **kwargs):
        if not self.is_fitted:
            raise RuntimeError(
                "The model hasn't been fitted yet, call .fit()."
            )
        return func(self, *args, **kwargs)

    return inner



"""
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from xarray import Dataset


@dataclass
class MetaData:
    id: str
    name: str
    latitude: float
    longitude: float


def get_covariates(
        location: str, start_date: str, end_date: str, variable: str,
        ) -> Dataset:
    """Return timeseries of covariate data."""
    raise NotImplementedError


def get_target(
        location: str, start_date: str, end_date: str, variabe: str
        ) -> Dataset:
    """Return target data."""
    raise NotImplementedError


def get_metadata(location: str) -> MetaData:
    """Return metadata."""
    raise NotImplementedError



import gpytorch
import numpy as np
import torch
from discontinuum.engines.gpytorch import MarginalGPyTorch
from gpytorch.kernels import (
    MaternKernel,
    PeriodicKernel,
    RBFKernel,
    ScaleKernel,
)
from gpytorch.priors import (
    GammaPrior,
    HalfNormalPrior,
    NormalPrior,
)

from loadest_gp.models.base import LoadestDataMixin, ModelConfig
from loadest_gp.plot import LoadestPlotMixin


class LoadestGPMarginalGPyTorch(
    LoadestDataMixin,
    LoadestPlotMixin,
    # comes last b/c it conflics with Mixin
    MarginalGPyTorch,
):
    """
    Gaussian Process implementation of the LOAD ESTimation (LOADEST) model

    This model currrently uses the marginal likelihood implementation, which is
    fast but does not account for censored data. Censored data require a slower
    latent variable implementation.
    """
    def __init__(
            self,
            model_config: ModelConfig = ModelConfig(),
    ):
        """ """
        super(MarginalGPyTorch, self).__init__(model_config=model_config)
        self.build_datamanager(model_config)

    def build_model(self, X, y) -> gpytorch.models.ExactGP:
        """Build marginal likelihood version of LoadestGP
        """
        # noise_prior = HalfNormalPrior(scale=0.01)
        # self.likelihood = gpytorch.likelihoods.GaussianLikelihood(
        #     noise_prior=noise_prior,
        # )

        noise = 0.1**2 * torch.ones(y.shape[0]).reshape(1, -1)
        self.likelihood = gpytorch.likelihoods.FixedNoiseGaussianLikelihood(
            noise=noise,
            learn_additional_noise=False,
        )

        model = ExactGPModel(X, y, self.likelihood)

        return model


class ExactGPModel(gpytorch.models.ExactGP):
    def __init__(self, train_x, train_y, likelihood):
        super(ExactGPModel, self).__init__(train_x, train_y, likelihood)

        n_d = train_x.shape[1]  # number of dimensions
        self.dims = np.arange(n_d)
        self.time_dim = [self.dims[0]]
        self.cov_dims = self.dims[1:]

        self.mean_module = gpytorch.means.ConstantMean()
        self.covar_module = (
            self.cov_trend()
            + self.cov_seasonal()
            + self.cov_covariates()
            + self.cov_residual()
        )

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)

    def cov_trend(self):
        eta = HalfNormalPrior(scale=1)
        ls = GammaPrior(concentration=4, rate=1)

        return ScaleKernel(
            RBFKernel(
                # active_dims=(0),
                active_dims=self.time_dim,
                lengthscale_prior=ls,
            ),
            outputscale_prior=eta,
        )

    def cov_seasonal(self):
        # TODO add lengthscale priors
        eta = HalfNormalPrior(scale=1)
        period = NormalPrior(loc=1, scale=0.01)

        return ScaleKernel(
            PeriodicKernel(
                period_length_prior=period,
                active_dims=self.time_dim,
                )
            * MaternKernel(
                nu=2.5,
                active_dims=self.time_dim
                ),
            outputscale_prior=eta,
        )

    def cov_covariates(self):
        eta = HalfNormalPrior(scale=2)
        ls = GammaPrior(concentration=2, rate=3)  # alpha, beta

        return ScaleKernel(
            RBFKernel(
                lengthscale_prior=ls,
                active_dims=self.cov_dims,
            ),
            outputscale_prior=eta,
        )

    def cov_residual(self):
        eta = HalfNormalPrior(scale=0.2)
        ls = GammaPrior(concentration=2, rate=10)

        return ScaleKernel(
            MaternKernel(
                ard_num_dims=2,
                nu=1.5,
                active_dims=self.dims,
                lengthscale_prior=ls,
            ),
            outputscale_prior=eta,
        )



"""Plotting functions"""

from __future__ import annotations

from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
from discontinuum.engines.base import is_fitted
from scipy.stats import norm
from xarray import DataArray
from xarray.plot.utils import label_from_attrs

if TYPE_CHECKING:
    from typing import Dict, Optional

    from matplotlib.pyplot import Axes
    from xarray import Dataset


DEFAULT_FIGSIZE = (5, 5)
NARROW_LINE = 1
REGULAR_LINE = NARROW_LINE * 1.5


class BasePlotMixin:
    """Mixin plotting functions for Model class"""

    @staticmethod
    def setup_plot(ax: Optional[Axes] = None):
        """Sets up figure and axes for plot.

        Parameters
        ----------
        ax : Axes, optional
            Pre-defined matplotlib axes.

        Returns
        -------
        ax : Axes
            Generated matplotlib axes.
        """
        if ax is None:
            _, ax = plt.subplots(1, figsize=DEFAULT_FIGSIZE)

        return ax

    @is_fitted
    def plot_observations(self, ax: Optional[Axes] = None, **kwargs):
        """Plot observations versus time.

        Parameters
        ----------
        ax : Axes, optional
            Pre-defined matplotlib axes.

        Returns
        -------
        ax : Axes
            Generated matplotlib axes.

        """
        # TODO provide easier access to data
        self.dm.data.target.plot.scatter(
            y=self.dm.data.target.name,
            c="k",
            s=5,
            linewidths=0.5,
            edgecolors="white",
            ax=ax,
            **kwargs,
        )

    @is_fitted
    def plot(self, covariates: Dataset, ci: float = 0.95, ax: Optional[Axes] = None):
        """Plot predicted data versus time.

        Parameters
        ----------
        covariates : Dataset
            Covariates.
        ci : float, optional
            Confidence interval. The default is 0.95.
        ax : Axes, optional
            Pre-defined matplotlib axes.

        Returns
        -------
        ax : Axes
            Generated matplotlib axes.
        """
        ax = self.setup_plot(ax)

        mu, se = self.predict(covariates, diag=True, pred_noise=True)

        target = DataArray(
            mu,
            coords=[covariates.time],
            dims=["time"],
            attrs=self.dm.data.target.attrs,
        )

        # compute confidence bounds
        lower, upper = self.dm.error_pipeline.ci(target, se, ci=ci)

        target.plot.line(ax=ax, lw=1, zorder=2)

        ax.fill_between(
            target["time"],
            lower,
            upper,
            color="b",
            alpha=0.1,
            zorder=1,
        )

        self.plot_observations(ax, zorder=3)

        return ax



# Retrieve data from the National Water Quality Assessment Program (NAWQA)

import boto3
import io
import lithops
import os
import pandas as pd
import xarray as xr

from dataclasses import dataclass
from matplotlib import pyplot as plt

from discontinuum.utils import aggregate_to_daily
from loadest_gp.providers.usgs import format_nwis_daily, format_wqp_samples
from loadest_gp import LoadestGPMarginalGPyTorch as LoadestGP

PROJECT = "National Water Quality Assessment Program (NAWQA)"
START_DATE = "1991-01-01"  # verify start date
END_DATE = "2024-12-31"  # verify end date
PCODE = "631"  # Nitrate plus nitrite
CHARACTERISTIC = "Inorganic nitrogen (nitrate and nitrite)"
FRACTION = "Dissolved"
DESTINATION_BUCKET = "wma-uncertainty"
DESTINATION_PATH = "nwqn-loadest-example/loadest-gp-output"
SAMPLES_BUCKET = "s3://wma-uncertainty/nwqn-loadest-example/nwqn-samples.parquet"
DAILY_BUCKET = "s3://wma-uncertainty/nwqn-loadest-example/nwqn-streamflow.parquet"


@dataclass
class SiteRecord:
    site_id: str
    start_date: str
    end_date: str
    characteristic: str
    fraction: str
    project: str


def map_retrieval(record: SiteRecord):
    """Map function to pull data from NWIS and WQP"""
    # download covariates (daily streamflow)
    #daily = usgs.get_daily(
    #    site=record.site_id,
    #    start_date=record.start_date,
    #    end_date=record.end_date
    #    )

    # download target (concentration)
    #samples = usgs.get_samples(
    #    site=record.site_id,
    #    start_date=record.start_date,
    #    end_date=record.end_date,
    #    characteristic=record.characteristic,
    #    fraction=record.fraction,
    #    project=record.project
    #    )

    # read data from s3
    site = record.site_id
    print(f"Processing site {site}")

    try:
        daily = pd.read_parquet(f"{DAILY_BUCKET}/site_no={site}")
        samples = pd.read_parquet(f"{SAMPLES_BUCKET}/MonitoringLocationIdentifier=USGS-{site}")
    except:
        print(f"Site {site} partion does not exist.")
        return

    samples = samples[samples["USGSPCode"] == PCODE]

    if samples.empty or daily.empty:
        print(f"Site {site} has no data.")
        return

    # convert parquet back to float
    # TODO: fix this upstream
    samples["ResultMeasureValue"] = samples["ResultMeasureValue"].astype(float)
    daily["00060_Mean"] = daily["00060_Mean"].astype(float)

    # check censoring threshold
    n = samples["ResultMeasureValue"].count()
    if n/len(samples) < 0.8:
        print(f"Site {site} has too many censored values.")
        return
    samples = format_wqp_samples(samples)
    samples = aggregate_to_daily(samples)

    daily = format_nwis_daily(daily)

    # TODO fix this upstream, data from other sites is being appended not updated
    daily = daily.drop_duplicates('time')
    daily = daily.dropna('time', how='any')

    training_data = xr.merge([samples, daily], join="inner")
    #training_data = training_data.dropna('time', how='any')

    min_obs = 50
    if len(training_data['time']) <= min_obs:
        print(f"Site {site} has fewer than {min_obs} observations.")
        return

    model = LoadestGP()

    try:
        model.fit(
            target=training_data["concentration"],
            covariates=training_data[["time", "flow"]]
            )

    except Exception as e:
        print(f"Site {site} failed to fit model: {e}")
        return

    # plot the result
    fig, ax = plt.subplots()
    model.plot(daily[["time", "flow"]], ax=ax)

    # save the figure to S3
    img_data = io.BytesIO()
    fig.savefig(img_data, format='png')
    img_data.seek(0)

    s3 = boto3.resource('s3')
    bucket = s3.Bucket(DESTINATION_BUCKET)
    bucket.put_object(
        Body=img_data,
        ContentType='image/png',
        Key="{path}/nwqn-{site_id}-{characteristic}-{fraction}.png".format(
            path=DESTINATION_PATH,
            **record.__dict__,
            )
    )


if __name__ == "__main__":
    project = "National Water Quality Assessment Program (NAWQA)"

    site_df = pd.read_csv(
        'NWQN_sites.csv',
        comment='#',
        dtype={'SITE_QW_ID': str, 'SITE_FLOW_ID': str},
        )

    site_list = site_df['SITE_QW_ID'].to_list()

    sites = [
        SiteRecord(
            site_id=site,
            start_date=START_DATE,
            end_date=END_DATE,
            characteristic=CHARACTERISTIC,
            fraction=FRACTION,
            project=PROJECT,
            )
        for site in site_list
    ]

    #sites = sites[:4]  # prune for testing

    fexec = lithops.FunctionExecutor(config_file="lithops.yaml")
    futures = fexec.map(map_retrieval, sites)

    futures.get_result()



from __future__ import annotations

import numpy as np
import pandas as pd


from abc import ABC, abstractmethod
from numpy.typing import ArrayLike
from scipy.stats import norm
from sklearn.base import BaseEstimator, OneToOneFeatureMixin, TransformerMixin
from sklearn.pipeline import Pipeline
from xarray import DataArray


def datetime_to_decimal_year(x: ArrayLike) -> ArrayLike:
    """Convert a timeseries to decimal year.

    Parameters
    ----------
    x : DataArray
        Timeseries to convert.

    Returns
    -------
    ArrayLike
        Decimal year array.
    """
    if not np.issubdtype(x.dtype, np.datetime64):
        raise ValueError("Array must contain numpy datetime64 objects.")

    # not flattening the array will cause an error
    dt = pd.to_datetime(x.flatten())

    julian = dt.to_julian_date()
    days_in_year = 365 + dt.is_leap_year
    start_of_year = pd.to_datetime(dt.year, format="%Y").to_julian_date()
    year = dt.year

    decimal_year = year + (julian - start_of_year)/(days_in_year)
    return decimal_year.to_numpy()


def decimal_year_to_datetime(x: ArrayLike) -> ArrayLike:
    """Convert a decimal year to a datetime.

    Parameters
    ----------
    x : ArrayLike
        Decimal year to convert.

    Returns
    -------
    ArrayLike
        Datetime array.
    """
    # x = x.flatten()
    year = np.floor(x).astype(int)
    remainder = x - year

    start_of_year = pd.to_datetime(year, format="%Y")

    # Calculate the number of days in the year
    days_in_year = 365 + start_of_year.is_leap_year

    # Calculate the datetime corresponding to the decimal year
    dt = start_of_year + pd.to_timedelta(remainder * days_in_year, unit="D")

    return dt.round("1s").to_numpy()


class BaseTransformer(TransformerMixin, BaseEstimator):
    """Base class for transformers."""
    def __init__(self):
        pass

    def fit(self, X, y=None):
        return self


class ClipTransformer(BaseTransformer):
    """Clip a variable."""
    def __init__(self, min: float = None, max: float = None):
        """Clip a variable.

        Parameters
        ----------
        min : float
            Minimum value to clip.

        max : float
            Maximum value to clip.
        """
        self.min = min
        self.max = max

    def transform(self, X):
        return np.clip(X, a_min=self.min, a_max=self.max)

    def inverse_transform(self, X):
        return self.transform(X)


class LogTransformer(BaseTransformer):
    """Log-transform a variable."""
    def transform(self, X):
        return np.log(X)

    def inverse_transform(self, X):
        return np.exp(X)


class ShapeTransformer(OneToOneFeatureMixin, BaseTransformer):
    """Reshape a 1D array to 2D.

    Reshaping is a persistent issue that is still handled poorly.
    """
    def transform(self, X):
        return X.reshape(-1, 1)

    def inverse_transform(self, X):
        return X.squeeze()


class SquareTransformer(OneToOneFeatureMixin, BaseTransformer):
    """Square a variable."""
    def transform(self, X):
        return X**2

    def inverse_transform(self, X):
        return np.sqrt(X)


class StandardScaler(BaseTransformer):
    """Rescale a variable to have a mean of 0 and a standard deviation of 1.

    Reimplemens the sklearn.preprocessing.StandardScaler but removes the
    requirement of having 2D arrays.
    """
    def __init__(self, *, with_mean=True, with_std=True):
        self.with_mean = with_mean
        self.with_std = with_std

    def fit(self, X, y=None):
        if self.with_mean:
            self.mean_ = X.mean(axis=0)
        if self.with_std:
            self.scale_ = X.std(axis=0)
        return self

    def transform(self, X):
        if self.with_mean:
            X = X - self.mean_
        if self.with_std:
            X = X / self.scale_
        return X

    def inverse_transform(self, X):
        if self.with_std:
            X = X * self.scale_
        if self.with_mean:
            X = X + self.mean_
        return X


class TimeTransformer(BaseTransformer):
    """Convert a datetime to decimal year."""
    def transform(self, X):
        return datetime_to_decimal_year(X)

    def inverse_transform(self, X):
        return decimal_year_to_datetime(X)


class MetadataManager(OneToOneFeatureMixin, TransformerMixin, BaseEstimator):
    def __init__(self):
        """Store xarray metadata (attrs)."""
        self.attrs = None

    def fit(self, X, y=None):
        """Store metadata from a xarray DataArray.

        Parameters
        ----------
        X : DataArray

        y : None
            Ignored.
        """
        self.attrs = X.attrs
        self.name = X.name
        self.dims = X.dims

        return self

    def transform(self, X):
        """
        TODO should we use a separate tranformer for reshaping?

        Parameters
        ----------
        X : DataArray

        Returns
        -------
        Numpy array
        """
        return X.values

    def inverse_transform(self, X):
        """Add xarray metadata to a numpy array.

        Parameters
        ----------
        X : Numpy array

        Returns
        -------
        DataArray
        """
        return DataArray(
            X,
            attrs=self.attrs,
            name=self.name,
            dims=self.dims,
        )


class LogStandardPipeline(Pipeline):
    def __init__(self):
        """Pipeline for log-distributed data.
        """
        super().__init__(
            steps=[
                ("metadata", MetadataManager()),
                ("reshape", ShapeTransformer()),
                ("clip", ClipTransformer(min=1e-6)),
                ("log", LogTransformer()),
                ("scaler", StandardScaler()),
            ]
        )


class NoOpPipeline(Pipeline):
    def __init__(self):
        """Pipeline that does not transform data.
        """
        super().__init__(
            steps=[
                ("metadata", MetadataManager()),
                ("reshape", ShapeTransformer()),
                ("clip", ClipTransformer(min=0)),
            ]
        )


class StandardPipeline(Pipeline):
    def __init__(self):
        """Pipeline for normally distributed data.
        """
        super().__init__(
            steps=[
                ("metadata", MetadataManager()),
                ("reshape", ShapeTransformer()),
                ("clip", ClipTransformer(min=0)),
                ("scaler", StandardScaler()),
            ]
        )


class TimePipeline(Pipeline):
    def __init__(self):
        """Pipeline for time data."""
        super().__init__(
            steps=[
                ("metadata", MetadataManager()),
                ("decimal_year", TimeTransformer()),
                ("reshape", ShapeTransformer()),
                ("scaler", StandardScaler(with_std=False)),
            ]
        )


class ErrorPipeline(Pipeline, ABC):
    @abstractmethod
    def ci(self, mean, se, ci=0.95):
        """Calculate confidence interval for a variable.

        Parameters
        ----------
        mean : float
            Mean of the variable.
        se : float
            Standard error of the variable.
        ci : float
            Confidence level.

        Returns
        -------
        lower, upper : Tuple[float, float]
            Lower and upper bound of the confidence interval.
        """
        pass


class StandardErrorPipeline(ErrorPipeline):
    """Pipeline to transform error

    inverse_transform converts variance to SE.
    """

    def __init__(self):
        super().__init__(
            steps=[
                ("metadata", MetadataManager()),
                ("reshape", ShapeTransformer()),
                ("scaler", StandardScaler(with_mean=False)),
                ("square", SquareTransformer()),
                # clip formerly used np.abs
                ("clip", ClipTransformer(min=0)),
            ]
        )

    def ci(self, mean, se, ci=0.95):
        """Calculate confidence interval for a standard variable.

        Parameters
        ----------
        mean : float
            Mean of the variable.
        se : float
            Standard error of the variable.
        ci : float
            Confidence level.

        Returns
        -------
        lower, upper : Tuple[float, float]
            Lower and upper bound of the confidence interval.
        """
        alpha = (1 - ci)/2
        zscore = norm.ppf(1-alpha)
        cb = se*zscore
        lower = mean - cb
        upper = mean + cb
        return lower, upper


class LogErrorPipeline(ErrorPipeline):
    """Pipelin to transform error

    inverse_transform converts variance (in log space) to a GSE
    """

    def __init__(self):
        super().__init__(
            steps=[
                ("metadata", MetadataManager()),
                ("reshape", ShapeTransformer()),
                ("log", LogTransformer()),
                ("scaler", StandardScaler(with_mean=False)),
                ("square", SquareTransformer()),
                # clip formerly used np.abs
                ("clip", ClipTransformer(min=1e-6)),
            ]
        )

    def ci(self, mean, se, ci=0.95):
        """Calculate confidence interval for a log-transformed variable.

        Parameters
        ----------
        mean : float
            Mean of the variable.
        se : float
            Standard error of the variable.
        ci : float
            Confidence level.

        Returns
        -------
        lower, upper : Tuple[float, float]
            Lower and upper bound of the confidence interval.
        """
        alpha = (1 - ci)/2
        zscore = norm.ppf(1-alpha)
        cb = se**zscore
        lower = mean / cb
        upper = mean * cb
        return lower, upper



"""Plotting functions"""

from __future__ import annotations

from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
from matplotlib.cm import ScalarMappable
import matplotlib.ticker as mticker
import numpy as np
import pandas as pd
from discontinuum.engines.base import is_fitted
from discontinuum.plot import BasePlotMixin
from scipy.stats import norm
import xarray as xr
from xarray import DataArray
from xarray.plot.utils import label_from_attrs

if TYPE_CHECKING:
    from typing import Dict, Optional, Union
    from datetime import datetime
    from matplotlib.pyplot import Axes
    from xarray import Dataset


DEFAULT_FIGSIZE = (5, 5)


class RatingPlotMixin(BasePlotMixin):
    """Mixin plotting functions for Model class"""

    @is_fitted
    def plot_observed_rating(self, ax: Optional[Axes] = None, **kwargs):
        """Plot stage observations versus discharge observations.

        Parameters
        ----------
        ax : Axes, optional
            Pre-defined matplotlib axes.

        Returns
        -------
        ax : Axes
            Generated matplotlib axes.
        """
        ax = self.setup_plot(ax)

        data = xr.merge([self.dm.data.covariates["stage"],
                         self.dm.data.target])

        # This converts the input uncertainty from GSE to SE
        se_unc = np.log(self.dm.data.target_unc) * self.dm.data.target
        ax.errorbar(self.dm.data.covariates["stage"],
                    self.dm.data.target,
                    yerr=se_unc,
                    c="k",
                    marker='',
                    linestyle='',
                    capsize=2.5)
        data.plot.scatter(x='stage',
                          y='discharge',
                          hue="time",
                          add_colorbar=False,
                          add_legend=False,
                          marker='o',
                          edgecolors='face',
                          s=10,
                          ax=ax,
                          **kwargs)

        return ax

    @is_fitted
    def plot_rating(self,
                    covariates: Dataset,
                    ci: float = 0.95,
                    ax: Optional[Axes] = None,
                    **kwargs
                   ):
        """Plot predicted discharge versus stage.

        Parameters
        ----------
        covariates : Dataset
            Covariates.
        ci : float, optional
            Confidence interval. The default is 0.95.
        ax : Axes, optional
            Pre-defined matplotlib axes.

        Returns
        -------
        ax : Axes
            Generated matplotlib axes.
        """
        ax = self.setup_plot(ax)

        mu, se = self.predict(covariates, diag=True, pred_noise=True)

        target = DataArray(
            mu,
            coords=[covariates.time],
            dims=["time"],
            attrs=self.dm.data.target.attrs,
        )

        # compute confidence bounds
        lower, upper = self.dm.error_pipeline.ci(target, se, ci=ci)

        ax.plot(covariates["stage"], target, lw=1, zorder=2, **kwargs)

        ax.fill_between(
            covariates['stage'],
            lower,
            upper,
            alpha=0.1,
            zorder=1,
            **kwargs
        )

        # self.plot_observed_rating(ax, zorder=3)

        return ax

    @is_fitted
    def plot_stage(self,
                   covariates: Dataset = None,
                   ax: Optional[Axes] = None,
                   **kwargs):
        """Plot observations versus time.

        Parameters
        ----------
        ax : Axes, optional
            Pre-defined matplotlib axes.

        Returns
        -------
        ax : Axes
            Generated matplotlib axes.

        """
        self.dm.data.covariates.plot.scatter(
            x='time',
            y='stage',
            hue='time',
            s=10,
            edgecolors="face",
            ax=ax,
            zorder=2,
            add_legend=False,
            add_colorbar=False,
            **kwargs,
        )

        if covariates is not None:
            covariates["stage"].plot.line(ax=ax, lw=1, zorder=1)

        return ax

    @is_fitted
    def plot_discharge(self,
                       covariates: Dataset = None,
                       ci: float = 0.95,
                       ax: Optional[Axes] = None
                      ):
        """Plot predicted discharge versus time.

        Parameters
        ----------
        covariates : Dataset, optional
            Covariates.
        ci : float, optional
            Confidence interval. The default is 0.95.
        ax : Axes, optional
            Pre-defined matplotlib axes.

        Returns
        -------
        ax : Axes
            Generated matplotlib axes.
        """
        ax = self.setup_plot(ax)

        if covariates is not None:
            mu, se = self.predict(covariates, diag=True, pred_noise=True)
    
            target = DataArray(
                mu,
                coords=[covariates.time],
                dims=["time"],
                attrs=self.dm.data.target.attrs,
            )
    
            # compute confidence bounds
            lower, upper = self.dm.error_pipeline.ci(target, se, ci=ci)

            target.plot.line(ax=ax, lw=1, zorder=2)

            ax.fill_between(
                target["time"],
                lower,
                upper,
                color="b",
                alpha=0.1,
                zorder=1,
            )

        self.dm.data.target.plot.scatter(
            x='time',
            y='discharge',
            hue='time',
            s=10,
            edgecolors="face",
            ax=ax,
            zorder=2,
            add_legend=False,
            add_colorbar=False,
        )

        return ax

    @is_fitted
    def add_time_colorbar(self,
                          ax: Optional[Axes] = None,
                          time_ticks: Optional[list[Union[str, datetime]]] = None,
                          **kwargs
                         ):
        """Add a colorbar of the data time range to the figure.

        Parameters
        ----------
        ax : Axes, optional
            Pre-defined matplotlib axes.
        time_ticks : list[str, datetime], optional
            List of times to place ticks on the colorbar.

        Returns
        -------
        cbar : Axes
            Generated matplotlib colorbar.
        """
        cbar_range = pd.to_datetime(
            [self.dm.data.covariates.time.min().values,
             self.dm.data.covariates.time.max().values]
        )
        if time_ticks is None:
            time_ticks = pd.date_range(*cbar_range, periods=5)

        if 'cmap' in kwargs:
            cmap = kwargs['cmap']
        else:
            cmap = None

        ticks = (time_ticks - cbar_range[0])/(cbar_range[1] - cbar_range[0])
        if not hasattr(ax, 'cbar'):
            cbar = plt.colorbar(ScalarMappable(cmap=cmap),
                                ax=ax,
                                aspect=30,
                                ticks=ticks,
                                format=mticker.FixedFormatter(
                                    time_ticks.strftime('%Y-%m-%d')
                                ),
                                **kwargs
                               )
        else:
            cbar = ax.cbar

        ax.cbar = cbar

        return cbar


    @is_fitted
    def plot_ratings_in_time(self,
                             time: Optional[list[Union[str, datetime]]] = None,
                             ci: Optional[int] = 0,
                             ax: Optional[Axes] = None,
                             **kwargs):
        """Plot predicted discharge versus stage.

        Parameters
        ----------
        time : list[str, datetime], optional
            List of times at which to plot the rating curve.
        ci : float, optional
            Confidence interval. Default is 0 (i.e., no confidence intervals).
        ax : Axes, optional
            Pre-defined matplotlib axes.

        Returns
        -------
        ax : Axes
            Generated matplotlib axes.
        """
        ax = self.setup_plot(ax)
        
        n = 250
        stage = np.linspace(self.dm.data.covariates['stage'].min()*0.95,
                            self.dm.data.covariates['stage'].max()*1.5,
                            n)
        
        if time is None:
            time = pd.date_range(
                self.dm.data.covariates['time'].min().values,
                self.dm.data.covariates['time'].max().values,
                periods=5
            )
        else:
            time = pd.to_datetime(time)
        
        cbar = self.add_time_colorbar(ax=ax, time_ticks=time, **kwargs)
        cbar_range = pd.to_datetime(
            [self.dm.data.covariates.time.min().values,
             self.dm.data.covariates.time.max().values]
        )

        for datetime in time:
            covariates = xr.Dataset(
                data_vars=dict(
                    stage=(["time"], stage),
                ),
                coords=dict(
                    time=np.repeat(datetime, n),
                ),
            )
            self.plot_rating(covariates,
                             ci=ci,
                             ax=ax,
                             color=cbar.cmap(
                                 ((datetime - cbar_range[0])
                                  /(cbar_range[1] - cbar_range[0]))
                             )
                             )

        return ax



"""Data transformations to improve optimization"""

from __future__ import annotations

from typing import TYPE_CHECKING

import gpytorch
import numpy as np
import torch
import tqdm
from xarray import DataArray

from discontinuum.engines.base import BaseModel, is_fitted

if TYPE_CHECKING:
    from numpy.typing import ArrayLike
    from typing import Dict, Optional, Tuple
    from xarray import Dataset


class NoOpMean(gpytorch.means.Mean):
    def forward(self, x):
        return x.squeeze(-1)


class LatentGPyTorch(BaseModel):
    def __init__(
        self,
        model_config: Optional[Dict] = None,
    ):
        """ """
        super().__init__(model_config=model_config)

    def fit(self, covariates, target=None):
        pass


class MarginalGPyTorch(BaseModel):
    def __init__(
        self,
        model_config: Optional[Dict] = None,
    ):
        """ """
        super(BaseModel, self).__init__(model_config=model_config)

    def fit(
            self,
            covariates: Dataset,
            target: Dataset,
            target_unc: Dataset = None,
            iterations: int = 100,
            optimizer: str = "adam",
            ):
        """Fit the model to data.

        Parameters
        ----------
        covariates : Dataset
            Covariates for training.
        target : Dataset
            Target data for training.
        target_unc : Dataset, optional
            Uncertainty on target data.
        iterations : int, optional
            Number of iterations for optimization. The default is 100.
        optimizer : str, optional
            Optimization method. The default is "adam".
        """
        self.is_fitted = True
        # setup data manager (self.dm)
        self.dm.fit(target=target, covariates=covariates, target_unc=target_unc)

        self.X = self.dm.X
        self.y = self.dm.y
        train_x = torch.tensor(self.X, dtype=torch.float32)
        train_y = torch.tensor(self.y, dtype=torch.float32)

        if target_unc is None:
            self.model = self.build_model(train_x, train_y)
            # also sets self.likelihood
        else:
            self.y_unc = self.dm.y_unc
            train_y_unc = torch.tensor(self.y_unc, dtype=torch.float32)
            self.model = self.build_model(train_x, train_y, train_y_unc)

        self.model.train()
        self.likelihood.train()

        # Use the adam optimizer
        if optimizer == "adam":
            optimizer = torch.optim.Adam(self.model.parameters(), lr=0.1)
        else:
            raise NotImplementedError("Only Adam optimizer is implemented")

        # "Loss" for GPs - the marginal log likelihood
        mll = gpytorch.mlls.ExactMarginalLogLikelihood(self.likelihood, self.model)

        pbar = tqdm.tqdm(range(iterations), ncols=70)
        for i in pbar:
            # Zero gradients from previous iteration
            optimizer.zero_grad()
            # Output from model
            output = self.model(train_x)
            # Calc loss and backprop gradients
            loss = -mll(output, train_y)
            loss.backward()
            pbar.set_postfix(loss=loss.item())
            optimizer.step()

    @is_fitted
    def predict(self,
                covariates: Dataset,
                diag=True,
                pred_noise=False,
                ) -> Tuple[DataArray, DataArray]:
        """Uses the fitted model to make predictions on new data.

        The input and output are in the original data space.

        Parameters
        ----------
        covariates : Dataset
            Covariates for prediction.
        diag : bool, optional
            Return only the diagonal of the covariance matrix.
            The default is True.
        pred_noise : bool, optional
            Include measurement uncertainty in the prediction.
            The default is False.

        Returns
        -------
        target : DataArray
            Target prediction.
        se : DataArray
            Standard error of the prediction.
        """
        Xnew = torch.tensor(
            self.dm.Xnew(covariates),
            dtype=torch.float32,
        )

        mu, var = self.__gpytorch_predict(Xnew)

        target = self.dm.y_t(mu)
        se = self.dm.error_pipeline.inverse_transform(var)

        return target, se

    @is_fitted
    def predict_grid(self, covariate: str, index="time", t_step=12):
        """Predict on a grid of points.

        Parameters
        ----------
        covariate_dim : int, optional
            Dimension to predict on. The default is 1.
        t_step : int, optional
            Time steps per year. The default is 12.
        """
        time_dim = self.dm.get_dim(index)
        covariate_dim = self.dm.get_dim(covariate)

        x_max = self.dm.X.max(axis=0)
        x_min = self.dm.X.min(axis=0)
        x_range = x_max - x_min

        n_cov = 18
        n_time = np.round(x_range[time_dim] * t_step).astype(int)

        x_time = torch.linspace(x_min[time_dim], x_max[time_dim], n_time)
        x_cov = torch.linspace(x_min[covariate_dim], x_max[covariate_dim], n_cov)

        # expects a 1D vector
        X_grid = torch.cartesian_prod(x_time, x_cov)

        mu, var = self.__gpytorch_predict(X_grid)

        target = self.dm.y_t(mu)
        target = target.data.reshape(n_time, n_cov)

        # TODO handle type conversion in pipeline
        index = self.dm.covariate_pipelines["time"].inverse_transform(x_time.numpy())
        covariate = self.dm.covariate_pipelines[covariate].inverse_transform(x_cov.numpy())

        return target, index, covariate

    @is_fitted
    def sample(self,
               covariates,
               n=1000,
               #diag=False,
               #pred_noise=False,
               #method="cholesky",
               #tol=1e-6,
               ) -> DataArray:
        """Sample from the posterior distribution of the model.

        Parameters
        ----------
        covariates : Dataset
            Covariates for prediction.
        n : int, optional
            Number of samples to draw.
        """
        Xnew = torch.tensor(
            self.dm.Xnew(covariates),
            dtype=torch.float32,
        )

        self.model.eval()
        self.likelihood.eval()

        f_preds = self.model(Xnew)

        sim = f_preds.sample(sample_shape=torch.Size([n]))

        # TODO modify transform to handle draws
        # flatten then reshape to work around our transformation pipeline
        temp = self.dm.y_t(sim.flatten())
        data = temp.data.reshape(n, -1)
        attrs = temp.attrs
        da = DataArray(
            data,
            coords={"time": covariates.time, "draw": np.arange(n)},
            dims=["draw", "time"],
            attrs=attrs,
        )

        return da

    def build_model(self, X, y, **kwargs) -> gpytorch.models.ExactGP:
        """
        Creates an instance of pm.Model based on provided data and
        model_config, and attaches it to self.

        The subclass method must instantiate self.model and self.likelihood.

        Raises
        ------
        NotImplementedError
        """
        self.model = None

        raise NotImplementedError(
            "This method must be implemented in a subclass"
            )

    @is_fitted
    def __gpytorch_predict(
            self,
            x: torch.Tensor,
            ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Model-space prediction.

        Parameters
        ----------
        x : torch.Tensor
            Input data in the (transformed) model space.

        Returns
        -------
        mu : torch.Tensor
            Mean of the prediction in the model space.
        var : torch.Tensor
            Variance of the prediction in the model space.
        """
        self.model.eval()
        self.likelihood.eval()

        with torch.no_grad(), gpytorch.settings.fast_pred_var():
            observed_pred = self.likelihood(self.model(x))
            mu = observed_pred.mean
            var = observed_pred.variance

        return mu, var



"""Data preprocessing utilities."""

from __future__ import annotations

import functools
from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING

import numpy as np
from sklearn.pipeline import Pipeline
from xarray import Dataset

from discontinuum.pipeline import (
    LogErrorPipeline,
    LogStandardPipeline,
    TimePipeline,
)

if TYPE_CHECKING:
    from typing import Dict, Type

    from numpy.typing import ArrayLike


def is_initialized(func):
    """Decorator checks whether model has been fit."""

    @functools.wraps(func)
    def inner(self, *args, **kwargs):
        if not self.is_initialized:
            raise RuntimeError(
                "The DataManager has not been initialized,",
                "call .init(target, covariates)."
            )
        return func(self, *args, **kwargs)

    return inner


@dataclass
class Data:
    target: Dataset
    covariates: Dataset
    target_unc: Dataset = None


# TODO create wrapper that validates input data
# TODO DataManager must track attrs
@dataclass
class DataManager:
    """ """

    target_pipeline: Type[Pipeline] = LogStandardPipeline
    error_pipeline: Type[Pipeline] = LogErrorPipeline
    covariate_pipelines: Dict[str, Pipeline] = None

    def fit(
            self,
            target: Dataset,
            covariates: Dataset,
            target_unc: Dataset = None):
        """Initialize DataManager for a given data distribution.

        Parameters
        ----------
        target : Dataset
            Target data.

        covariates : Dataset
            Covariate data.

        target_unc : Dataset
            Target uncertainty. Default is None.
        """
        # ensure time comes first in the dict

        default_pipeline = {"time": TimePipeline}
        default_pipeline.update(self.covariate_pipelines)
        self.covariate_pipelines = default_pipeline

        self.data = Data(target, covariates, target_unc)

        self.target_pipeline = self.target_pipeline().fit(target)
        self.error_pipeline = self.error_pipeline().fit(target)

        for key, value in self.covariate_pipelines.items():
            self.covariate_pipelines[key] = value().fit(covariates[key])

    def transform_covariates(self, covariates: Dataset) -> ArrayLike:
        """Transform covariates into design matrix"""
        X = np.empty((len(covariates.time), len(self.covariate_pipelines)))
        for i, (key, value) in enumerate(self.covariate_pipelines.items()):
            X[:, i] = value.transform(covariates[key]).flatten()
        return X

    # TODO handle reshaping in pipeline
    def inverse_transform_covariates(self, X: ArrayLike) -> Dataset:
        """Inverse transform design matrix into covariates"""
        # inverse transform each column of X using the corresponding pipeline
        covariates = {}
        for i, (key, value) in enumerate(self.covariate_pipelines.items()):
            covariates[key] = value.inverse_transform(X[:, i])
        return Dataset(covariates)

    @cached_property
    def y(self, dtype="float32") -> ArrayLike:
        """Convenience function for DataManager.target.transform"""
        return self.target_pipeline.transform(self.data.target).flatten()

    @cached_property
    def y_unc(self, dtype="float32") -> ArrayLike:
        """Convenience function for DataManager.target.transform"""
        return self.error_pipeline.transform(self.data.target_unc).flatten()

    @cached_property
    def X(self) -> ArrayLike:
        """Convenience function for DataManager.covariates.transform"""
        return self.transform_covariates(self.data.covariates)

    def Xnew(self, ds: Dataset) -> ArrayLike:
        """Convenience function for DataManager.covariates.transform"""
        return self.transform_covariates(ds)

    def y_t(self, y: ArrayLike) -> Dataset:
        """Convenience function for DataManager.target.untransform"""
        return self.target_pipeline.inverse_transform(y)

    def get_dim(self, dim: str, index="time") -> int:
        """Get the dimension of a variable.

        In other words, its column in the design matrix.

        Parameters
        ----------
        dim : str
            Dimension name.

        Returns
        -------
        int
            Dimension (column) in design matrix.
        """
        cov_list = [index] + list(self.data.covariates)

        return cov_list.index(dim)



from __future__ import annotations

from typing import TYPE_CHECKING
from warnings import warn

import matplotlib.pyplot as plt
import numpy as np

if TYPE_CHECKING:
    from typing import Optional

    from matplotlib.pyplot import Axes
    from xarray import DataArray


def concentration_to_flux(
        concentration: DataArray,
        flow: DataArray,
        ) -> DataArray:
    """Convert concentration (mg/l) to flux (kg).

    Parameters
    ----------
    concentration : DataArray
        Concentration data.
    flow : DataArray
        Flow data.

    Returns
    -------
    DataArray
        Flux data.
    """
    time_delta = np.unique(
        concentration.time.diff(dim="time").dt.total_seconds()
    )
    mg_l_to_kg_m3 = 1e-3

    if len(time_delta) != 1:
        warn("Time delta is not constant",
             stacklevel=UserWarning,)

    if flow.units != "cubic meters per second":
        warn(
            "Check that flow is 'cubic meters per second'.",
            "Set flow.units = 'cubic meters per second' to silence.",
            stacklevel=UserWarning,
            )

    if "mg/l" not in concentration.units:
        warn("Check that concentration is in 'mg/l'.",
             "Set concentration.units = 'mg/l' to silence.",
             stacklevel=UserWarning,
             )

    flux = concentration * flow * time_delta * mg_l_to_kg_m3
    flux.attrs = concentration.attrs
    flux.attrs["units"] = "kilograms"
    flux.attrs["standard_name"] = "flux"
    return flux


def plot_annual_flux(
        flux: DataArray,
        ax: Optional[Axes] = None,
        ) -> Axes:
    """Plot annual flux.

    Parameters
    ----------
    flux : DataArray
        Flux data.
    ax : Axes, optional
        Pre-defined matplotlib axes.

    Returns
    -------
    ax : Axes
        Generated matplotlib axes.
    """
    _, ax = plt.subplots() if ax is None else (ax.figure, ax)

    annual = flux.resample(time="YE").sum()
    annual.attrs["units"] = "kilograms per year"

    annual_df = annual.to_dataframe(name=annual.attrs["standard_name"])
    annual_df.boxplot(by="time", showfliers=False, grid=False, ax=ax)

    ax.set_ylabel(
        "{}\n[{}]".format(annual.attrs["long_name"], annual.attrs["units"])
        )
    ax.set_xlabel("Year")
    ax.tick_params(axis="x", labelrotation=90)

    years = annual.time.dt.year.values
    labels = ["" if (year % 5 != 0) else year for year in years]
    ax.set_xticklabels(labels)
    ax.set_title("")
    return ax
